# =============================================================
#
# Copyright 2021-2023, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# =============================================================

# AUTHORS:
# - B-Open Solutions srl

"""
This test checks that JPEG-LS compressed MTG FCI L1C products are correctly decompressed using the
netCDF4 tool ``nccopy`` by comparing bands in the decompressed file with the same bands in
a reference output.
"""

import os
import subprocess

import netCDF4 as nc
import numpy as np
import pytest


TEST_DATA_PATH = os.environ.get("EPCT_TEST_DATA_DIR", "")
INPUT_PATH = os.path.join(TEST_DATA_PATH, "MTG", "MTGFCIL1")
COMP_FILEPATH = [os.path.join(INPUT_PATH, file_name) for file_name in (
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY--"
    "DIS-NC4E_C_EUMT_20200405000845_GTT_DEV_20200405000330_20200405000345_N_JLS_T_0001_0015.nc",
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY--"
    "DIS-NC4E_C_EUMT_20200405120015_GTT_DEV_20200405115500_20200405115515_N_JLS_T_0072_0021.nc"
)]
DECOMP_FILEPATH = [os.path.join(INPUT_PATH, file_name) for file_name in (
    "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY---"
    "NC4E_C_EUMT_20200405000845_GTT_DEV_20200405000330_20200405000345_N__T_0001_0015.nc",
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY---"
    "NC4E_C_EUMT_20200405120015_GTT_DEV_20200405115500_20200405115515_N__T_0072_0021.nc"
)]
BANDS = [
    "ir_105", "ir_123", "ir_133", "ir_38", "ir_87", "ir_97",
    "nir_13", "nir_16", "nir_22",
    "vis_04", "vis_05", "vis_06", "vis_08", "vis_09",
    "wv_63", "wv_73"
]
VARIABLES = [
    "effective_radiance", "pixel_quality", "index_map"
]


def test_decompression(tmpdir):

    assert "HDF5_PLUGIN_PATH" in os.environ.keys()

    for (comp_test_file, decomp_test_file) in zip(COMP_FILEPATH, DECOMP_FILEPATH):

        decomp_res_file = os.path.join(tmpdir, os.path.basename(decomp_test_file))
        process = subprocess.run(
            f"nccopy -F none {comp_test_file} {decomp_res_file}", shell=True
        )
        decomp_file_size = os.path.getsize(decomp_res_file)
        comp_file_size = os.path.getsize(comp_test_file)

        assert os.path.isfile(decomp_res_file)
        assert decomp_file_size > (comp_file_size * 4)

        ds_test = nc.Dataset(decomp_test_file, "r")
        ds_res = nc.Dataset(decomp_res_file, "r")

        for band in BANDS:
            for variable in VARIABLES:
                array_test = ds_test[f"data/{band}/measured/{variable}"][:]
                array_res = ds_res[f"data/{band}/measured/{variable}"][:]
                assert np.ma.allequal(array_test, array_res)





# =============================================================
#
# Copyright 2021-2023, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# =============================================================

# AUTHORS:
# - B-Open Solutions srl

"""
This test measures speed performances of JPEG-LS compressed MTG FCI L1C products decompression
through the `netCDF4` Python library.
"""

import os

import netCDF4 as nc
import numpy as np
import pytest

import fcidecomp


TEST_DATA_PATH = os.environ.get("EPCT_TEST_DATA_DIR", "")
INPUT_PATH = os.path.join(TEST_DATA_PATH, "MTG", "MTGFCIL1")
COMP_FILEPATH = [os.path.join(INPUT_PATH, file_name) for file_name in (
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY--"
    "DIS-NC4E_C_EUMT_20200405000845_GTT_DEV_20200405000330_20200405000345_N_JLS_T_0001_0015.nc",
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY--"
    "DIS-NC4E_C_EUMT_20200405120015_GTT_DEV_20200405115500_20200405115515_N_JLS_T_0072_0021.nc"
)]
DECOMP_FILEPATH = [os.path.join(INPUT_PATH, file_name) for file_name in (
    "W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY---"
    "NC4E_C_EUMT_20200405000845_GTT_DEV_20200405000330_20200405000345_N__T_0001_0015.nc",
    "W_XX-EUMETSAT-Darmstadt_IMG+SAT_MTI1+FCI-1C-RRAD-FDHSI-FD--CHK-BODY---"
    "NC4E_C_EUMT_20200405120015_GTT_DEV_20200405115500_20200405115515_N__T_0072_0021.nc"
)]


def py_decompression(comp_file):

    ds = nc.Dataset(comp_file, "r")
    band = ds['data']['vis_04']['measured']['effective_radiance'][:]
    ds.close()

    return band


@pytest.mark.parametrize(
    "test_input,expected",
    [(test, exp) for (test, exp) in zip(COMP_FILEPATH, DECOMP_FILEPATH)],
    ids=[f"{os.path.getsize(comp_file) / (1024 * 1024):.2f} Mb" for comp_file in COMP_FILEPATH]
)
def test_decomp_py(benchmark, test_input, expected):

    assert "HDF5_PLUGIN_PATH" in os.environ.keys()

    band_test = benchmark(py_decompression, test_input)
    ds_exp = nc.Dataset(expected, 'r')
    band_exp = ds_exp['data']['vis_04']['measured']['effective_radiance'][:]

    assert np.ma.allequal(band_test, band_exp)




# Copyright 2021, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os

# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import re

project = "EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan"
copyright = (
    "2021, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)"
)
author = "B-Open Solutions srl"

# The full version, including alpha/beta/rc tags.
release = re.sub("^v", "", os.popen("git describe").read().strip())
if "CI_COMMIT_SHORT_SHA" in os.environ:
    release = os.environ.get("CI_COMMIT_SHORT_SHA")
if "CI_COMMIT_TAG" in os.environ:
    release = os.environ.get("CI_COMMIT_TAG")

version = release


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"font_family": "Barlow", "head_font_family": "Dosis ExtraLight"}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}
html_logo = "images/eumetsat.png"

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "user_guidedoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    "papersize": "a4paper",
    "maketitle": r"""
        \pagenumbering{Roman} %%% to avoid page 1 conflict with actual page 1

        \sphinxmaketitle

        \clearpage
        \pagenumbering{arabic}

        """,
    "preamble": r"""
        \usepackage{graphicx}
        \usepackage{background}
    
        \backgroundsetup{
          scale=1,
          color=black,
          opacity=1,
          angle=0,
          position=current page.north,
          contents={%
          \small\sffamily%
          \begin{minipage}{.22\textwidth}
          \vspace{1.65cm}
          \hspace{-0.175cm}
          \includegraphics[width=\linewidth,height=70pt,keepaspectratio]{../../images/eumetsat.png}
          \end{minipage}%
          \begin{minipage}{.8\textwidth}
          \vspace{2cm}
          \parbox[b]{.6\textwidth}{}\hfill \\
          \end{minipage}%
          }
        }
    
        \usepackage{fancyhdr}
        \pagestyle{fancy}
        \fancypagestyle{normal}{%
        \fancyhead{}
        \fancyhead[RE,RO]{\bf{documentation/verification-validation-plan \\ version, \today \\ EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan}}
        \renewcommand{\headrulewidth}{0.5pt}
        \fancyfoot{}
        \fancyfoot[C]{\thepage}
        }
        \fancypagestyle{plain}{%
        \fancyhead{}
        \fancyhead[RE,RO]{\bf{documentation/verification-validation-plan \\ version, \today \\ EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan}}
        \fancyfoot[CO,CE]{\thepage}
        }

        """.replace("version", version),
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "fcidecomp-verification-validation-plan.tex",
        "EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan",
        "EUMETSAT",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "fcidecomp-verification-validation-plan",
        "EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan",
        [author],
        1,
    )
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "fcidecomp-verification-validation-plan",
        "EUMETSAT Data Tailor Web Service - Support to CharLS decompression for MTG users - Verification and Validation Plan",
        author,
        "fcidecomp-verification-validation-plan",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

numfig = True

# -- Extension configuration -------------------------------------------------
def setup(app):
    app.add_css_file("custom.css")



# =============================================================
#
# Copyright 2021-2023, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# =============================================================

# AUTHORS:
# - B-Open Solutions srl

import ctypes as _ctypes
from glob import glob as _glob
import logging as _logging
import os as _os
import sys as _sys
from collections.abc import Mapping as _Mapping
import h5py as _h5py

_logger = _logging.getLogger(__name__)

try:
    PLUGINS_PATH = _os.environ["HDF5_PLUGIN_PATH"]
except KeyError:
    print("HDF5_PLUGIN_PATH environment variable is not defined")

FILTER_NAME = 'fcidecomp'
FILTER_ID = 32018

try:
    _FilterRefClass = _h5py.filters.FilterRefBase
except AttributeError:
    class _FilterRefClass(_Mapping):
        """
        Base class for referring to an HDF5 filter and describing its options.
        """
        filter_id = None
        filter_options = ()

        # Mapping interface supports using instances as **kwargs for compatibility
        # with older versions of h5py
        @property
        def _kwargs(self):
            return {
                'compression': self.filter_id,
                'compression_opts': self.filter_options
            }

        def __hash__(self):
            return hash((self.filter_id, self.filter_options))

        def __len__(self):
            return len(self._kwargs)

        def __iter__(self):
            return iter(self._kwargs)

        def __getitem__(self, item):
            return self._kwargs[item]


class FciDecomp(_FilterRefClass):
    """`
    `h5py.Group.create_dataset``'s compression arguments for using FciDecomp filter.
    """
    filter_id = FILTER_ID

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _init_filters():
    """
    Initialise and register HDF5 filters with h5py
    Generator of tuples: (filename, library handle)
    """
    hdf5_version = _h5py.h5.get_libversion()

    filter_name = FILTER_NAME
    filter_id = FILTER_ID

    # Check if filter is already loaded (not on buggy HDF5 versions)
    if (1, 8, 20) <= hdf5_version < (1, 10) or hdf5_version >= (1, 10, 2):
        if _h5py.h5z.filter_avail(filter_id):
            _logger.warning("%s filter already loaded, skip it.", filter_name)
            return

    # Load DLL
    if _sys.platform.startswith('win'):
        filter_file_name = 'H5Zjpegls.dll'
    elif _sys.platform.startswith('linux'):
        filter_file_name = 'libH5Zjpegls.so'
    else:
        filter_file_name = ''
    filename = _glob(_os.path.join(PLUGINS_PATH, filter_file_name))
    if len(filename):
        filename = filename[0]
    else:
        _logger.error("Cannot initialize filter %s: File not found", filter_name)
        return
    try:
        lib = _ctypes.CDLL(filename)
    except OSError:
        _logger.error("Failed to load filter %s: %s", filter_name, filename)
        return

    if _sys.platform.startswith('win'):
        # Use register_filter function to register filter
        lib.register_filter.restype = _ctypes.c_int
        retval = lib.register_filter()
    else:
        # Use init_filter function to initialize DLL and register filter
        lib.init_filter.argtypes = [_ctypes.c_char_p]
        lib.init_filter.restype = _ctypes.c_int
        retval = lib.init_filter(
            bytes(_h5py.h5z.__file__, encoding='utf-8'))

    if retval < 0:
        _logger.error("Cannot initialize filter %s: %d", filter_name, retval)
        return

    yield filename, lib


_filters = dict(_init_filters())  # Store loaded filters



# =============================================================
#
# Copyright 2021-2023, European Organisation for the Exploitation of Meteorological Satellites (EUMETSAT)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# =============================================================

# AUTHORS:
# - B-Open Solutions srl

import os
from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

install_requirements = ["h5py"]
setup_requirements = ["setuptools"]

try:
    with open("VERSION") as version_file:
        version = version_file.read().strip()
except FileNotFoundError:
    version = os.environ.get(
        "CI_COMMIT_TAG", os.popen("git describe --tags --abbrev=0").read()
        )
    if version == "":
        version = "0.0.0+"+os.environ.get("CI_COMMIT_SHORT_SHA")
    with open("VERSION", "w") as version_file:
        version_file.write(version)

setup(
    name='fcidecomp-python',
    version=version,
    description="FCIDECOMP Python plugin",
    long_description=readme,
    author="B-Open Solutions srl",
    author_email="software@bopen.eu",
    url="https://www.eumetsat.int",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requirements,
    setup_requires=setup_requirements,
    license="Apache License 2.0'",
    zip_safe=True,
    keywords="",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)
